//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/SampleDesigner/ParticleCoreShellForm.cpp
//! @brief     Implements class ParticleCoreShellForm
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/SampleDesigner/ParticleCoreShellForm.h"
#include "GUI/Model/Sample/FormFactorItems.h"
#include "GUI/Model/Sample/ParticleCoreShellItem.h"
#include "GUI/Model/Sample/ParticleItem.h"
#include "GUI/Model/Types/VectorDescriptor.h"
#include "GUI/View/SampleDesigner/ActionFactory.h"
#include "GUI/View/SampleDesigner/FormLayouter.h"
#include "GUI/View/SampleDesigner/SampleEditorController.h"
#include "GUI/View/SampleDesigner/SelectionContainerForm.h"
#include "GUI/View/Tool/GroupBoxCollapser.h"

#include <QAction>
#include <QComboBox>

namespace {

QComboBox* createFormFactorCombo(QWidget* parent, FormFactorItem* current)
{
    auto* combo = new QComboBox(parent);
    WheelEventEater::install(combo);
    combo->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Fixed);

    combo->addItem(QIcon(), "None", "");
    for (auto modelType : ItemCatalog::formFactorTypes()) {
        const auto ui = ItemCatalog::instance().uiInfo(modelType);
        combo->addItem(QIcon(ui.iconPath), ui.menuEntry, modelType);
    }
    combo->setMaxVisibleItems(combo->count());

    if (current == nullptr)
        combo->setCurrentIndex(0);
    else
        combo->setCurrentIndex(combo->findData(current->modelType()));

    return combo;
}

} // namespace


ParticleCoreShellForm::ParticleCoreShellForm(QWidget* parent, ParticleCoreShellItem* item,
                                             SampleEditorController* ec, bool allowRemove)
    : QGroupBox(parent), m_item(item), m_ec(ec)
{
    setTitle("Core/shell particle");
    FormLayouter layouter(this, ec);
    layouter.setContentsMargins(30, 6, 0, 0);
    layouter.addVector(item->positionVector(), false);
    layouter.addSelection(item->rotationMethod());
    layouter.addValue(item->abundance());

    // - core
    {
        QGroupBox* coreParticleGroup = new QGroupBox(this);
        coreParticleGroup->setObjectName("Particle");

        core.layouter.reset(new FormLayouter(coreParticleGroup, ec));

        core.formFactorCombo = createFormFactorCombo(
            coreParticleGroup, item->core() != nullptr ? item->core()->formFactor() : nullptr);
        connect(core.formFactorCombo, QOverload<int>::of(&QComboBox::currentIndexChanged), this,
                &ParticleCoreShellForm::onCoreComboChanged);
        core.layouter->addRow("Form factor:", core.formFactorCombo);
        core.collapser = GroupBoxCollapser::installIntoGroupBox(coreParticleGroup);

        auto* showInRealSpaceAction =
            ActionFactory::createShowInRealSpaceAction(this, "core particle");
        connect(showInRealSpaceAction, &QAction::triggered, this,
                &ParticleCoreShellForm::showCoreInRealSpace);
        core.collapser->addAction(showInRealSpaceAction);

        createCoreWidgets();

        layouter.addRow(coreParticleGroup);
    }

    // - shell
    {
        QGroupBox* shellParticleGroup = new QGroupBox(this);
        shellParticleGroup->setObjectName("Particle");
        shell.layouter.reset(new FormLayouter(shellParticleGroup, ec));
        shell.formFactorCombo = createFormFactorCombo(
            shellParticleGroup, item->shell() != nullptr ? item->shell()->formFactor() : nullptr);
        connect(shell.formFactorCombo, QOverload<int>::of(&QComboBox::currentIndexChanged), this,
                &ParticleCoreShellForm::onShellComboChanged);
        shell.layouter->addRow("Form factor:", shell.formFactorCombo);
        shell.collapser = GroupBoxCollapser::installIntoGroupBox(shellParticleGroup);

        auto* showInRealSpaceAction =
            ActionFactory::createShowInRealSpaceAction(this, "shell particle");
        connect(showInRealSpaceAction, &QAction::triggered, this,
                &ParticleCoreShellForm::showShellInRealSpace);
        shell.collapser->addAction(showInRealSpaceAction);

        createShellWidgets();

        layouter.addRow(shellParticleGroup);
    }

    auto* mainCollapser = GroupBoxCollapser::installIntoGroupBox(this);

    auto* showInRealSpaceAction = ActionFactory::createShowInRealSpaceAction(
        this, "core/shell particle", [=] { ec->requestViewInRealSpace(item); });

    mainCollapser->addAction(showInRealSpaceAction);

    if (allowRemove) {
        m_removeAction = ActionFactory::createRemoveAction(this, "core/shell particle",
                                                           [=] { ec->removeParticle(item); });
        mainCollapser->addAction(m_removeAction);
    }
}

void ParticleCoreShellForm::onCoreComboChanged()
{
    while (core.layouter->layout()->rowCount() > 1)
        core.layouter->layout()->removeRow(1);

    const QString newClassName = core.formFactorCombo->currentData().toString();
    m_ec->setCoreFormFactor(this, newClassName);
}

void ParticleCoreShellForm::onShellComboChanged()
{
    while (shell.layouter->layout()->rowCount() > 1)
        shell.layouter->layout()->removeRow(1);

    const QString newClassName = shell.formFactorCombo->currentData().toString();
    m_ec->setShellFormFactor(this, newClassName);
}

void ParticleCoreShellForm::showCoreInRealSpace()
{
    if (ParticleItem* particle = dynamic_cast<ParticleItem*>(m_item->core()))
        m_ec->requestViewInRealSpace(particle);
}

void ParticleCoreShellForm::showShellInRealSpace()
{
    if (ParticleItem* particle = dynamic_cast<ParticleItem*>(m_item->shell()))
        m_ec->requestViewInRealSpace(particle);
}

void ParticleCoreShellForm::createCoreWidgets()
{
    QString groupTitle = "Core";

    if (ParticleItem* particle = dynamic_cast<ParticleItem*>(m_item->core())) {
        const QString formFactor =
            particle->formFactorItem()->value().value<ComboProperty>().getValue();

        groupTitle += " (" + formFactor + ")";

        core.layouter->addGroupOfValues("Geometry", particle->formFactor()->geometryValues());
        core.layouter->addVector(particle->positionVector(), false);
        core.layouter->addSelection(particle->rotationMethod());
        core.layouter->addValue(particle->abundance());
    }

    core.collapser->setTitle(groupTitle);
}

void ParticleCoreShellForm::createShellWidgets()
{
    QString groupTitle = "Shell";

    if (ParticleItem* particle = dynamic_cast<ParticleItem*>(m_item->shell())) {
        const QString formFactor =
            particle->formFactorItem()->value().value<ComboProperty>().getValue();

        groupTitle += " (" + formFactor + ")";

        shell.layouter->addGroupOfValues("Geometry", particle->formFactor()->geometryValues());
        shell.layouter->addVector(particle->positionVector(), false);
        shell.layouter->addSelection(particle->rotationMethod());
        shell.layouter->addValue(particle->abundance());
    }

    shell.collapser->setTitle(groupTitle);
}

void ParticleCoreShellForm::enableStructureEditing(bool b)
{
    if (m_removeAction)
        m_removeAction->setVisible(b);
}

ParticleCoreShellItem* ParticleCoreShellForm::coreShellItem() const
{
    return m_item;
}
